@@ -8,7 +8,7 @@ import warnings
 from struct import calcsize, pack, unpack
 
 from pfio import cache
-from pfio.cache.file_cache import _DEFAULT_CACHE_PATH
+from pfio.cache.file_cache import _check_local, _default_cache_path
 
 
 class _NoOpenNamedTemporaryFile(object):
@@ -125,6 +125,7 @@ class MultiprocessFileCache(cache.Cache):
 
         dir (str): The path to the directory to place cache data in
             case home directory is not backed by fast storage device.
+            Must not be an NFS.
 
         cache_size_limit (None or int): Limitation of the cache size in bytes.
             If the total amount of cached data reaches the limit,
@@ -155,10 +156,11 @@ class MultiprocessFileCache(cache.Cache):
         self.cache_size_limit = cache_size_limit
 
         if dir is None:
-            self.dir = _DEFAULT_CACHE_PATH
+            self.dir = _default_cache_path()
         else:
             self.dir = dir
         os.makedirs(self.dir, exist_ok=True)
+        _check_local(self.dir)
 
         self.closed = False
         self._frozen = False
