@@ -1,14 +1,20 @@
 import abc
+import configparser
 import contextlib
 import copy
 import os
 import stat
+import warnings
 from abc import abstractmethod
 from io import IOBase
 from types import TracebackType
-from typing import Any, Callable, Iterator, Optional, Type
+from typing import Any, Callable, Iterator, Optional, Type, Union
 from urllib.parse import urlparse
 
+from deprecation import deprecated
+
+from pfio.version import __version__  # NOQA
+
 
 class FileStat(abc.ABC):
     """Detailed file or directory information abstraction
@@ -42,7 +48,7 @@ class FileStat(abc.ABC):
     size = None
 
     def isdir(self):
-        """Returns whether the target is a directory, based on the permission flag
+        """Returns whether the target is a directory from the permission flag
 
         Note that some systems do not support directory tree semantics.
 
@@ -103,9 +109,9 @@ class FS(abc.ABC):
                  [str, int], Any]] = None) -> Type["IOBase"]:
         raise NotImplementedError()
 
-    def open_zip(self, file_path: str, mode='r') -> Type["Zip"]:  # NOQA
+    def open_zip(self, file_path: str, mode='r', **kwargs) -> Type["Zip"]:  # NOQA
         from .zip import Zip
-        return Zip(self, file_path, mode)
+        return Zip(self, file_path, mode, **kwargs)
 
     def subfs(self, rel_path: str) -> Type["FS"]:
         '''Virtually changes the working directory
@@ -124,11 +130,20 @@ class FS(abc.ABC):
         sub = copy.copy(self)
 
         sub._cwd = os.path.join(self.cwd, rel_path)
+        sub._reset()
         return sub
 
     def _checkfork(self):
-        if self.is_forked:
-            raise ForkedError()
+        if not self.is_forked:
+            return
+
+        # Forked!
+        self._reset()
+        self.pid = os.getpid()
+
+    @abstractmethod
+    def _reset(self):
+        raise NotImplementedError()
 
     @property
     def is_forked(self):
@@ -140,7 +155,7 @@ class FS(abc.ABC):
 
     @abstractmethod
     def list(self, path_or_prefix: Optional[str] = None,
-             recursive=False) -> Iterator[str]:
+             recursive=False, detail=False) -> Iterator[Union[FileStat, str]]:
         """Lists all the files and directories under
            the given ``path_or_prefix``
 
@@ -155,6 +170,9 @@ class FS(abc.ABC):
             recursive (bool): When this is ``True``, list files and directories
                 recursively.
 
+            detail (bool): If this is ``True``, the return values will be the
+                detail information of each file or directory.
+
         Returns:
             An Iterator that iterates though the files and directories.
 
@@ -306,6 +324,9 @@ def from_url(url: str, **kwargs) -> 'FS':
     built from it accordingly. The URL path is supposed to
     be a directory for file systems or a path prefix for S3.
 
+    .. warning:: When opening an ``hdfs://...`` URL, be sure about
+        forking context. See: :class:`Hdfs` for discussion.
+
     Arguments:
         url (str): A URL string compliant with RFC 1738.
 
@@ -318,16 +339,25 @@ def from_url(url: str, **kwargs) -> 'FS':
     .. note:: Some FS resouces won't be closed when using this
         functionality.
 
-    '''
-    parsed = urlparse(url)
-    force_type = kwargs.pop('force_type', None)
+    .. note:: Pickling the FS object may or may not work correctly
+        depending on the implementation.
 
+    '''
+    if kwargs.pop('reset_on_fork', None) is not None:
+        warnings.warn(
+            "reset_on_fork is deprecated. PFIO resets on fork by default",
+            category=DeprecationWarning,
+            stacklevel=2
+        )
+
+    parsed = urlparse(url)
     if parsed.scheme:
         scheme = parsed.scheme
     else:
         scheme = 'file'  # Default is local
 
     # When ``force_type`` is defined, it must be equal with given one.
+    force_type = kwargs.pop('force_type', None)
     if force_type is not None and force_type != "zip":
         if force_type != scheme:
             raise ValueError("URL scheme mismatch with forced type")
@@ -346,14 +376,14 @@ def from_url(url: str, **kwargs) -> 'FS':
         _zip_check_create_not_supported()
         dirname, filename = os.path.split(parsed.path)
         fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)
-        fs = fs.open_zip(filename)
+        fs = fs.open_zip(filename, **kwargs)
 
     elif force_type is None:
         if parsed.path.endswith('.zip'):
             _zip_check_create_not_supported()
             dirname, filename = os.path.split(parsed.path)
             fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)
-            fs = fs.open_zip(filename)
+            fs = fs.open_zip(filename, **kwargs)
         else:
             dirname = parsed.path
             fs = _from_scheme(scheme, dirname, kwargs, bucket=parsed.netloc)
@@ -365,7 +395,54 @@ def from_url(url: str, **kwargs) -> 'FS':
     return fs
 
 
+def _default_config_file():
+    path = os.getenv('PFIO_CONFIG_PATH')
+    if path:
+        return path
+
+    basedir = os.getenv('XDG_CONFIG_HOME')
+    if not basedir:
+        basedir = os.path.join(os.getenv('HOME'), ".config")
+
+    return os.path.join(basedir, "pfio.ini")
+
+
+class _CustomScheme:
+    conf = None
+
+    @staticmethod
+    def config(scheme):
+        if _CustomScheme.conf is None:
+            _CustomScheme.load_config()
+
+        if scheme in _CustomScheme.conf:
+            return dict(_CustomScheme.conf[scheme])
+
+    @staticmethod
+    def load_config():
+        config = configparser.ConfigParser()
+        configfile = _default_config_file()
+        config.read(configfile)
+        _CustomScheme.conf = config
+
+
 def _from_scheme(scheme, dirname, kwargs, bucket=None):
+    known_scheme = ['file', 'hdfs', 's3']
+
+    # Custom scheme; using configparser for older Python. Will
+    # update to toml in Python 3.11 once 3.10 is in the end.
+    if scheme not in known_scheme:
+        config_dict = _CustomScheme.config(scheme)
+        if config_dict is not None:
+            scheme = config_dict.pop('scheme')  # Get the real scheme
+            # Custom scheme expected here
+            if scheme not in known_scheme:
+                raise ValueError("Scheme {} is not supported", scheme)
+            for k in config_dict:
+                if k not in kwargs:
+                    # Don't overwrite with configuration value
+                    kwargs[k] = config_dict[k]
+
     if scheme == 'file':
         from .local import Local
         fs = Local(dirname, **kwargs)
@@ -376,11 +453,13 @@ def _from_scheme(scheme, dirname, kwargs, bucket=None):
         from .s3 import S3
         fs = S3(bucket=bucket, prefix=dirname, **kwargs)
     else:
-        raise RuntimeError("Scheme {} is not supported", scheme)
+        raise RuntimeError("bug: scheme '{}' is not supported".format(scheme))
 
     return fs
 
 
+@deprecated(deprecated_in='2.2.0', removed_in='2.3.0',
+            current_version=__version__)
 def lazify(init_func, lazy_init=True, recreate_on_fork=True) -> "FS":
     '''Make FS init lazy and recreate on fork
 
