@@ -1,15 +1,17 @@
 import getpass
 import io
 import logging
+import multiprocessing
 import os
 import re
 import subprocess
+import warnings
 from xml.etree import ElementTree
 
 import pyarrow
 from pyarrow.fs import FileSelector, FileType, HadoopFileSystem
 
-from .fs import FS, FileStat
+from .fs import FS, FileStat, ForkedError
 
 logger = logging.getLogger(__name__)
 logger.addHandler(logging.StreamHandler())
@@ -178,6 +180,28 @@ class Hdfs(FS):
     be defined instead. ``$CLASSPATH`` will be needed in case ``hdfs``
     command is not available from ``$PATH``.
 
+
+    .. warning:: It is strongly discouraged to use :class:`Hdfs` under
+          multiprocessing. Once the object detects the process id changed
+          (which means it is forked), the object raises :class:`ForkedError`
+          before doing anything. If you do *need* forking, for example, PyTorch
+          DataLoader with multiple workers for performance, it is strongly
+          recommended not to instantiate :class:`Hdfs` before forking. Details
+          are described in PFIO issue #123.  Simple workaround is to set
+          multiprocessing start method as ``'forkserver'`` and start
+          the very first child process before everything.
+
+          .. code-block::
+
+             import multiprocessing
+             multiprocessing.set_start_method('forkserver')
+             p = multiprocessing.Process()
+             p.start()
+             p.join()
+
+
+          See: https://github.com/pfnet/pfio/issues/123
+
     .. note:: With environment variable
           ``KRB5_KTNAME=path/to/your.keytab`` set, ``hdfs``
           handler automatically starts automatic and periodical
@@ -187,6 +211,7 @@ class Hdfs(FS):
     .. note::
           Only the username in the first entry in The
           keytab will be used to update the Kerberos ticket.
+
     '''
 
     def __init__(self, cwd=None, create=False, **_):
@@ -205,10 +230,30 @@ class Hdfs(FS):
 
         if not self.isdir(''):
             if create:
-                self.makedirs('')
+                # Since this process (isdir -> makedirs) is not atomic,
+                # makedirs can conflict in case of a parallel workload.
+                self.makedirs('', exist_ok=True)
             else:
                 raise ValueError('{} must be a directory'.format(self.cwd))
 
+        if multiprocessing.get_start_method() != 'forkserver':
+            # See https://github.com/pfnet/pfio/pull/123 for detail
+            warnings.warn('Non-forkserver start method under HDFS detected.')
+
+    def _reset(self):
+        if multiprocessing.get_start_method() != 'forkserver':
+            raise ForkedError()
+
+        self._fs = _create_fs()
+
+    def __getstate__(self):
+        state = self.__dict__.copy()
+        state['_fs'] = None
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__ = state
+
     def _get_principal_name(self):
         # get the default principal name from `klist` cache
         principal_name = _get_principal_name_from_klist()
@@ -261,7 +306,7 @@ class Hdfs(FS):
     def close(self):
         pass
 
-    def list(self, path: str = "", recursive=False):
+    def list(self, path: str = "", recursive=False, detail=False):
         self._checkfork()
 
         if not self.isdir(path):
@@ -272,7 +317,10 @@ class Hdfs(FS):
 
         infos = self._fs.get_file_info(FileSelector(path, recursive=recursive))
         for file_info in infos:
-            yield file_info.path[len(norm_path)+1:]
+            if detail:
+                yield HdfsFileStat(file_info)
+            else:
+                yield file_info.path[len(norm_path)+1:]
 
     def stat(self, path):
         self._checkfork()
