@@ -53,7 +53,9 @@ class Local(FS):
 
         if not self.isdir(''):
             if create:
-                os.makedirs(self._cwd)
+                # Since this process (isdir -> makedirs) is not atomic,
+                # makedirs can conflict in case of a parallel workload.
+                os.makedirs(self._cwd, exist_ok=True)
             else:
                 raise ValueError('{} must be a directory'.format(self._cwd))
 
@@ -64,6 +66,9 @@ class Local(FS):
 
         return os.getcwd()
 
+    def _reset(self):
+        pass
+
     def open(self, file_path, mode='r',
              buffering=-1, encoding=None, errors=None,
              newline=None, closefd=True, opener=None):
@@ -73,31 +78,40 @@ class Local(FS):
                        buffering, encoding, errors,
                        newline, closefd, opener)
 
-    def list(self, path_or_prefix: str = '', recursive=False):
+    def list(self, path_or_prefix: str = '', recursive=False,
+             detail=False):
         path_or_prefix = os.path.join(self.cwd, path_or_prefix)
 
         if recursive:
             path_or_prefix = path_or_prefix.rstrip("/")
             # plus 1 to include the trailing slash
             prefix_end_index = len(path_or_prefix) + 1
-            yield from self._recursive_list(prefix_end_index, path_or_prefix)
+            yield from self._recursive_list(prefix_end_index,
+                                            path_or_prefix, detail)
         else:
             for e in os.scandir(path_or_prefix):
                 # ls -F
-                if e.is_dir():
+                if detail:
+                    yield LocalFileStat(e.stat(), e.name)
+                elif e.is_dir():
                     yield e.name + '/'
                 else:
                     yield e.name
 
-    def _recursive_list(self, prefix_end_index: int, path: str):
-        for file in os.scandir(path):
+    def _recursive_list(self, prefix_end_index: int, path: str,
+                        detail: bool):
+        for e in os.scandir(path):
             # ls -F
-            if file.is_dir():
-                yield file.path[prefix_end_index:] + '/'
-                yield from self._recursive_list(prefix_end_index,
-                                                file.path)
+            if detail:
+                yield LocalFileStat(e.stat(), e.name)
+            elif e.is_dir():
+                yield e.path[prefix_end_index:] + '/'
             else:
-                yield file.path[prefix_end_index:]
+                yield e.path[prefix_end_index:]
+
+            if e.is_dir():
+                yield from self._recursive_list(prefix_end_index,
+                                                e.path, detail)
 
     def stat(self, path):
         path = os.path.join(self.cwd, path)
