@@ -24,8 +24,8 @@ class S3ObjectStat(FileStat):
     def __init__(self, key, head):
         self.filename = key
         self.last_modified = head['LastModified'].timestamp()
-        self.size = head['ContentLength']
-        self.metadata = head['Metadata']
+        self.size = head.get('ContentLength', head.get('Size'))
+        self.metadata = head.get('Metadata', {})
         self._head = head
 
     def isdir(self):
@@ -307,9 +307,11 @@ class S3(FS):
                  mpu_chunksize=32*1024*1024,
                  buffering=-1,
                  create=False,
+                 _skip_connect=None,  # For test purpose
                  **_):
         super().__init__()
         self.bucket = bucket
+        self.create_bucket = create_bucket
         if prefix is not None:
             self.cwd = prefix
         else:
@@ -333,10 +335,11 @@ class S3(FS):
         # AWS_SECRET_ACCESS_KEY.
         self.aws_access_key_id = aws_access_key_id
         if aws_access_key_id is not None:
-            kwargs['aws_access_key_id'] = aws_access_key_id
+            kwargs['aws_access_key_id'] = os.path.expandvars(aws_access_key_id)
         self.aws_secret_access_key = aws_secret_access_key
         if aws_secret_access_key is not None:
-            kwargs['aws_secret_access_key'] = aws_secret_access_key
+            kwargs['aws_secret_access_key'] = os.path.expandvars(
+                aws_secret_access_key)
 
         # We won't expect any enviroment variable for S3 endpoints
         # supported by boto3. Instead, we take S3_ENDPOINT in case
@@ -352,23 +355,43 @@ class S3(FS):
         if self.endpoint is not None:
             kwargs['endpoint_url'] = self.endpoint
 
+        self.kwargs = kwargs
+        if not _skip_connect:
+            self._connect()
+
+    def _reset(self):
+        self._connect()
+
+    def _connect(self):
         # print('boto3.client options:', kwargs)
-        self.client = boto3.client('s3', **kwargs)
+        self.client = boto3.client('s3', **self.kwargs)
 
         try:
-            self.client.head_bucket(Bucket=bucket)
+            self.client.head_bucket(Bucket=self.bucket)
         except ClientError as e:
-            if e.response['Error']['Code'] == '404' and create_bucket:
-                res = self.client.create_bucket(Bucket=bucket)
-                print("Bucket", bucket, "created:", res)
+            if e.response['Error']['Code'] == '404' and self.create_bucket:
+                res = self.client.create_bucket(Bucket=self.bucket)
+                print("Bucket", self.bucket, "created:", res)
             else:
                 raise e
 
+    def __getstate__(self):
+        state = self.__dict__.copy()
+        state['client'] = None
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__ = state
+
     def open(self, path, mode='r', **kwargs):
         '''Opens an object accessor for read or write
 
         .. note:: Multi-part upload is not yet available.
 
+        Arguments:
+            path (str): relative path from basedir
+
+            mode (str): open mode
         '''
         self._checkfork()
         if 'a' in mode:
@@ -380,14 +403,24 @@ class S3(FS):
         path = _normalize_key(path)
         if 'r' in mode:
             obj = _ObjectReader(self.client, self.bucket, path, mode, kwargs)
+
+            bs = self.buffering
+            if bs < 0:
+                bs = min(obj.content_length, DEFAULT_MAX_BUFFER_SIZE)
+
             if 'b' in mode:
-                if self.buffering:
-                    bs = self.buffering
-                    if bs < 0:
-                        bs = min(obj.content_length, DEFAULT_MAX_BUFFER_SIZE)
+                if self.buffering and bs != 0:
                     obj = io.BufferedReader(obj, buffer_size=bs)
             else:
                 obj = io.TextIOWrapper(obj)
+                if self.buffering:
+                    # This is undocumented property; but resident at
+                    # least since 2009 (the merge of io-c branch).
+                    # We'll use it until the day of removal.
+                    if bs == 0:
+                        # empty file case: _CHUNK_SIZE must be positive
+                        bs = DEFAULT_MAX_BUFFER_SIZE
+                    obj._CHUNK_SIZE = bs
 
         elif 'w' in mode:
             obj = _ObjectWriter(self.client, self.bucket, path, mode,
@@ -400,7 +433,7 @@ class S3(FS):
 
         return obj
 
-    def list(self, prefix: str = "", recursive=False):
+    def list(self, prefix: str = "", recursive=False, detail=False):
         '''List all objects (and prefixes)
 
         Although there is not concept of directory in AWS S3 API,
@@ -430,9 +463,15 @@ class S3(FS):
         for res in iterator:
             # print(res)
             for common_prefix in res.get('CommonPrefixes', []):
-                yield common_prefix['Prefix'][len(key):]
+                if detail:
+                    yield S3PrefixStat(common_prefix['Prefix'][len(key):])
+                else:
+                    yield common_prefix['Prefix'][len(key):]
             for content in res.get('Contents', []):
-                yield content['Key'][len(key):]
+                if detail:
+                    yield S3ObjectStat(content['Key'][len(key):], content)
+                else:
+                    yield content['Key'][len(key):]
 
     def stat(self, path):
         '''Imitate FileStat with S3 Object metadata
