@@ -1,11 +1,11 @@
 import io
 import logging
 import os
-import sys
-import warnings
 import zipfile
 from datetime import datetime
 
+from pfio.cache.sparse_file import MPCachedWrapper
+
 from .fs import FS, FileStat
 
 logger = logging.getLogger(__name__)
@@ -51,12 +51,19 @@ class ZipFileStat(FileStat):
 
 class Zip(FS):
     _readonly = True
+    '''
+            local_cache (bool): use sparse file cache for opening ZIP file
 
-    def __init__(self, backend, file_path, mode='r', create=False, **_):
+            local_cachedir (dir): local path to store sparse file cache
+    '''
+
+    def __init__(self, backend, file_path, mode='r', create=False,
+                 local_cache=False, local_cachedir=None, **kwargs):
         super().__init__()
         self.backend = backend
         self.file_path = file_path
         self.mode = mode
+        self.kwargs = kwargs
 
         if create:
             raise ValueError("create option is not supported")
@@ -67,32 +74,59 @@ class Zip(FS):
         if 'w' in mode:
             self._readonly = False
 
-        self.fileobj = self.backend.open(file_path, mode + 'b')
+        self.local_cache = local_cache
+        self.local_cachedir = local_cachedir
+        self.local_cachefile = None
+        self.local_indexfile = None
 
-        if isinstance(self.backend, Zip) \
-           and sys.version_info < (3, 7, ):
-            # In Python < 3.7, the returned file object from zipfile.open,
-            # i.e. ZipExtFile, is not seekable,
-            # while in order to open as zip, the zipfile module requires
-            # the given file object to be seekable, which makes
-            # nested zip impossible.
-            # As a workaround, in case of nested zip,  we read the
-            # whole nested zipfile into BytesIO object,
-            # which is a seekable file object, upon open.
-            # However, it might cause performance and memory
-            # issues when the zipfile is huge. A warning is generated
-            # for user.
+        self._reset()
 
-            warnings.warn('In Python < 3.7, '
-                          'To support opening nested zip as container, '
-                          'PFIO has to read '
-                          'the entire nested zip upon open, '
-                          'which might cause performance or '
-                          'memory issues when the nested zip is huge.',
-                          category=RuntimeWarning)
-            self.fileobj = io.BytesIO(self.fileobj.read())
+    def _reset(self):
+        buffering = self.kwargs.get('buffering', 16*1024*1024)
 
-        self.zipobj = zipfile.ZipFile(self.fileobj, mode)
+        if self.local_cache:
+            # Don't use io.BufferedReader for sparse file cache
+            self.kwargs['buffering'] = 0
+
+        obj = self.backend.open(self.file_path,
+                                self.mode + 'b',
+                                **self.kwargs)
+
+        if 'w' not in self.mode:
+            stat = self.backend.stat(self.file_path)
+
+        # Use sparse file cache: Optimization for a remote object
+        # store system e.g. AWS S3 or HDFS
+        if self.local_cache:
+            self.kwargs['buffering'] = buffering
+
+            obj = MPCachedWrapper(obj, stat.size, self.local_cachedir,
+                                  local_cachefile=self.local_cachefile,
+                                  local_indexfile=self.local_indexfile,
+                                  close_on_close=True,
+                                  multithread_safe=True)
+
+            # Update local cachefile in case of being forked
+            self.local_cachefile = obj.local_cachefile
+            self.local_indexfile = obj.local_indexfile
+
+            # Default 16MB buffer size
+            if buffering > 0:
+                obj = io.BufferedReader(obj, buffer_size=buffering)
+
+        self.fileobj = obj
+
+        assert self.fileobj is not None
+        self.zipobj = zipfile.ZipFile(self.fileobj, self.mode)
+
+    def __getstate__(self):
+        state = self.__dict__.copy()
+        state['fileobj'] = None
+        state['zipobj'] = None
+        return state
+
+    def __setstate__(self, state):
+        self.__dict__ = state
 
     def open(self, file_path, mode='r',
              buffering=-1, encoding=None, errors=None,
@@ -132,7 +166,8 @@ class Zip(FS):
 
         return ZipFileStat(self.zipobj.getinfo(actual_path))
 
-    def list(self, path_or_prefix: str = "", recursive=False):
+    def list(self, path_or_prefix: str = "", recursive=False,
+             detail=False):
         self._checkfork()
 
         if path_or_prefix:
@@ -159,14 +194,19 @@ class Zip(FS):
                     "{} is not found".format(path_or_prefix))
 
         if recursive:
-            for name in self.zipobj.namelist():
+            for info in self.zipobj.infolist():
+                name = info.filename
                 if name.startswith(path_or_prefix):
                     name = name[len(path_or_prefix):].strip("/")
                     if name:
-                        yield name
+                        if detail:
+                            yield ZipFileStat(info)
+                        else:
+                            yield name
         else:
             _list = set()
-            for name in self.zipobj.namelist():
+            for info in self.zipobj.infolist():
+                name = info.filename
                 return_file_name = None
                 current_dir_list = os.path.normpath(name).split('/')
                 if not given_dir_list:
@@ -183,7 +223,10 @@ class Zip(FS):
                 if (return_file_name is not None
                         and return_file_name not in _list):
                     _list.add(return_file_name)
-                    yield return_file_name
+                    if detail:
+                        yield ZipFileStat(info)
+                    else:
+                        yield return_file_name
 
     def isdir(self, file_path: str):
         self._checkfork()
